/*
	$Id: AnnotationInRestriction.java 6880 2015-04-21 06:26:57Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.Query;
import net.sf.basedb.util.EqualsHelper;

import org.hibernate.SQLQuery;
import java.util.Arrays;
import java.util.List;

/**
	Restricts a query using annotation values that exists in a set
	of given values.
	
	@author Nicklas
	@version 2.0
	@see net.sf.basedb.core.query.Annotations
	@base.modified $Date: 2015-04-21 08:26:57 +0200 (ti, 21 apr 2015) $
*/
public class AnnotationInRestriction
	extends AnnotationRestriction
{
	private final List<?> values;

	/**
		@since 2.4
		@deprecated In 3.5; use {@link #AnnotationInRestriction(String, AnnotationType, boolean, boolean, Object...)} instead
	*/
	@Deprecated
	public AnnotationInRestriction(String alias, AnnotationType annotationType, boolean includeInheriting, Object... values)
		throws InvalidDataException
	{
		this(alias, annotationType, true, includeInheriting, values);
	}
	
	/**
		Create a new annotation restriction. At least one of
		includePrimary and includeInheriting must be true.
	 	
	 	@param alias The alias to use in the restriction.		
		@param annotationType The annotation type to use in the query
		@param includePrimary If items with primary annotation should be returned or not
		@param includeInheriting If items inherting the annotation should be returned
			by the query or not
		@param values An array of values, each value must be of the 
			correct value type for the annotation as defined by the 
			{@link AnnotationType#getValueType()} property
		@throws InvalidDataException If any of the parameters are null
			or the array is empty or not follow the rules above.
		@since 3.5
	*/
	public AnnotationInRestriction(String alias, AnnotationType annotationType, boolean includePrimary, boolean includeInheriting, Object... values)
		throws InvalidDataException
	{
		super(alias, annotationType, includePrimary, includeInheriting);
		if (values == null || values.length == 0) throw new InvalidUseOfNullException("values");
		for (int i = 0; i < values.length; ++i)
		{
			Object value = values[i];
			if (!valueType.isCorrectType(value))
			{
				throw new InvalidDataException("Value["+i+"] '"+value+"' is a "+
					value.getClass()+", not a "+valueType);
			}
		}
		this.values = Arrays.asList(values);
	}
	
	/*
		From the AnnotationRestriction class
		-------------------------------------------
	*/
	@Override
	String getRestrictionSql(Query query, DbControl dc)
		throws BaseException
	{
		org.hibernate.dialect.Dialect dialect = HibernateUtil.getDialect();
		return "v."+dialect.openQuote()+"value"+dialect.closeQuote()+" in (:values)";
	}
	
	@Override
	void setRestrictionParameters(SQLQuery query)
	{
		query.setParameterList("values", values);
	}
	// -----------------------------------------
	/*
		From the Object class
		-------------------------------------------
	*/
	/**
		This restriction is equal to another AnnotationInRestriction if they 
		have the same annotation type, list of values and inheritance setting.
	*/
	@Override
	public boolean equals(Object other)
	{
		Boolean eq = maybeEquals(other);
		if (eq != null) return eq;
		AnnotationInRestriction o = (AnnotationInRestriction)other;
		return EqualsHelper.invariantEquals(values.toArray(), o.values.toArray());
	}
	@Override
	public int hashCode()
	{
		return super.hashCode() + EqualsHelper.hashCode(values.toArray());
	}
	// -------------------------------------------

}

