/*
	$Id: BatchUtil.java 6875 2015-04-20 09:08:18Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.BatchableData;

import java.sql.Types;
import java.util.LinkedList;
import java.util.Map;
import java.util.List;
import java.util.HashMap;
import java.util.Iterator;

import org.hibernate.mapping.Table;
import org.hibernate.mapping.PersistentClass;
import org.hibernate.mapping.Column;
import org.hibernate.mapping.Property;
import org.hibernate.dialect.Dialect;

/**
	Utility class for batching information.

	@author Samuel, Nicklas
	@version 2.0
	@base.modified $Date: 2015-04-20 11:08:18 +0200 (må, 20 apr 2015) $
*/
class BatchUtil
{
	/**
		Log all SQL statements.
	*/
	private static final org.slf4j.Logger logSql = 
		org.slf4j.LoggerFactory.getLogger("net.sf.basedb.core.batcher.sql");

	/**
		So we don't always have to call logSql.debug()
	*/
	private static final boolean debugEnabled = logSql.isDebugEnabled();

	/**
		Mapping classes/entity names to property insert orders.
	*/
	private static Map<String, List<BatchedPropertyInfo>> propertyInsertOrders = null;
	
	/**
		Mapping classes/entity names to property delete orders.
	*/
	private static Map<String, List<BatchedPropertyInfo>> propertyUpdateOrders = null;
	
	/**
		Mapping classes/entity names to property update orders.
	*/
	private static Map<String, List<BatchedPropertyInfo>> propertyDeleteOrders = null;

	/**
		Mapping classes/entity names to SQL inserts.
	*/
	private static Map<String, String> sqlInserts = null;
	
	/**
		Mapping classes/entity names to SQL updates.
	*/
	private static Map<String, String> sqlUpdates = null;
	
	/**
		Mapping classes/entity names to SQL deletes.
	*/
	private static Map<String, String> sqlDeletes = null;
	
	private static boolean isInitialised = false;
	
	/**
		Generate SQL for all BatchableData classes for insert,
		update and delete operations. Keep track of the order
		of parameters and properties in the generated SQL.
	*/
	static synchronized void init()
	{
		if (isInitialised) return;
		propertyInsertOrders = new HashMap<String, List<BatchedPropertyInfo>>();
		propertyUpdateOrders = new HashMap<String, List<BatchedPropertyInfo>>();
		propertyDeleteOrders = new HashMap<String, List<BatchedPropertyInfo>>();
		sqlInserts = new HashMap<String, String>();
		sqlUpdates = new HashMap<String, String>();
		sqlDeletes = new HashMap<String, String>();

		Dialect dialect = HibernateUtil.getDialect();
		Iterator<PersistentClass> iterator = HibernateUtil.getClassMappings();
		while (iterator.hasNext())
		{
			PersistentClass pc = iterator.next();
			Class<?> c = pc.getMappedClass();
			if (BatchableData.class.isAssignableFrom(c))
			{
				buildInsert(pc.getEntityName(), dialect);
				buildUpdate(pc.getEntityName(), dialect);
				buildDelete(pc.getEntityName(), dialect);
			}
		}
		isInitialised = true;
	}
	
	/**
		Unload all settings.
	*/
	static synchronized void unload()
	{
		isInitialised = false;
		if (propertyInsertOrders != null) propertyInsertOrders.clear();
		propertyInsertOrders = null;
		if (propertyUpdateOrders != null) propertyUpdateOrders.clear();
		propertyUpdateOrders = null;
		if (propertyDeleteOrders != null) propertyDeleteOrders.clear();
		propertyDeleteOrders = null;
		if (sqlInserts != null) sqlInserts.clear();
		sqlInserts = null;
		if (sqlUpdates != null) sqlUpdates.clear();
		sqlUpdates = null;
		if (sqlDeletes != null) sqlDeletes.clear();
		sqlDeletes = null;
	}

	/**
		Get the SQL that inserts a new row of the specified class.
		@param dataClass Get the insert SQL for this class
		@return The SQL string
	*/
	static String getInsertSql(Class<? extends BatchableData> dataClass)
	{
		assert sqlInserts != null : "BatchUtil is not initiated.";
		assert dataClass != null : "dataClass == null";
		return sqlInserts.get(dataClass.getName());
	}
	
	/**
		Get the SQL that inserts a new row of the specified class.
		@param entityName The entity name
		@return The SQL string
	*/
	static String getInsertSql(String entityName)
	{
		assert sqlInserts != null : "BatchUtil is not initiated.";
		assert entityName != null : "entityName == null";
		return sqlInserts.get(entityName);
	}
	
	/**
		Get the SQL that updates a new row of the specified class.
		@param dataClass Get the update SQL for this class
		@return The SQL string
	*/
	static String getUpdateSql(Class<? extends BatchableData> dataClass)
	{
		assert sqlUpdates != null : "BatchUtil is not initiated.";
		assert dataClass != null : "dataClass == null";
		return sqlUpdates.get(dataClass.getName());
	}

	/**
		Get the SQL that updates a new row of the specified class.
		@param entityName The entity name
		@return The SQL string
	*/
	static String getUpdateSql(String entityName)
	{
		assert sqlUpdates != null : "BatchUtil is not initiated.";
		assert entityName != null : "entityName == null";
		return sqlUpdates.get(entityName);
	}
	
	/**
		Get the SQL that deletes a new row of the specified class.
		@param dataClass Get the delete SQL for this class
		@return The SQL string
	*/
	static String getDeleteSql(Class<? extends BatchableData> dataClass)
	{
		assert sqlDeletes != null : "BatchUtil is not initiated.";
		assert dataClass != null : "dataClass == null";
		return sqlDeletes.get(dataClass.getName());
	}
	
	/**
		Get the SQL that deletes a new row of the specified class.
		@param entityName The entity name
		@return The SQL string
	*/
	static String getDeleteSql(String entityName)
	{
		assert sqlDeletes != null : "BatchUtil is not initiated.";
		assert entityName != null : "entityName == null";
		return sqlDeletes.get(entityName);
	}
	
	/**
		Get the order of properties in the insert SQL.
		@param dataClass Get the property order for this class
		@return List with the properties in order
	*/
	static List<BatchedPropertyInfo> getInsertPropertyOrder(Class<? extends BatchableData> dataClass)
	{
		assert propertyInsertOrders != null : "BatchUtil is not initiated.";
		assert dataClass != null : "dataClass == null";
		return propertyInsertOrders.get(dataClass.getName());
	}

	/**
		Get the order of properties in the insert SQL.
		@param entityName The entity name
		@return List with the properties in order
	*/
	static List<BatchedPropertyInfo> getInsertPropertyOrder(String entityName)
	{
		assert propertyInsertOrders != null : "BatchUtil is not initiated.";
		assert entityName != null : "entityName == null";
		return propertyInsertOrders.get(entityName);
	}
	
	/**
		Get the order of properties in the update SQL.
		@param dataClass Get the property order for this class.
		@return List with the properties in order.
	*/
	static List<BatchedPropertyInfo> getUpdatePropertyOrder(Class<? extends BatchableData> dataClass)
	{
		assert propertyUpdateOrders != null : "BatchUtil is not initiated.";
		assert dataClass != null : "dataClass == null";
		return propertyUpdateOrders.get(dataClass.getName());
	}

	/**
		Get the order of properties in the update SQL.
		@param entityName The entity name
		@return List with the properties in order.
	*/
	static List<BatchedPropertyInfo> getUpdatePropertyOrder(String entityName)
	{
		assert propertyUpdateOrders != null : "BatchUtil is not initiated.";
		assert entityName != null : "entityName == null";
		return propertyUpdateOrders.get(entityName);
	}
	
	/**
		Get the order of properties in the delete SQL.
		@param dataClass Get the property order for this class.
		@return List with the properties in order.
	*/
	static List<BatchedPropertyInfo> getDeletePropertyOrder(Class<? extends BatchableData> dataClass)
	{
		assert propertyDeleteOrders != null : "BatchUtil is not initiated.";
		assert dataClass != null : "dataClass == null";
		return propertyDeleteOrders.get(dataClass.getName());
	}

	/**
		Get the order of properties in the delete SQL.
		@param entityName The entity name
		@return List with the properties in order.
	*/
	static List<BatchedPropertyInfo> getDeletePropertyOrder(String entityName)
	{
		assert propertyDeleteOrders != null : "BatchUtil is not initiated.";
		assert entityName != null : "entityName == null";
		return propertyDeleteOrders.get(entityName);
	}
	
	static String buildInsertSql(VirtualDb vdb, VirtualTable table, VirtualColumn[] columns, Object[] values)
	{
		Dialect dialect = HibernateUtil.getDialect();
		char oq = dialect.openQuote();
		char cq = dialect.closeQuote();
		
		StringBuilder sql = new StringBuilder();
		StringBuilder valuesSql = new StringBuilder();
		sql.append("INSERT INTO ");
		sql.append(table.getQualifiedTableName(vdb));	
		sql.append(" (");
		String deliminator = "";
		for (int i = 0; i < columns.length; ++i)
		{
			sql.append(deliminator).append(oq).append(columns[i].getColumn()).append(cq);
			valuesSql.append(deliminator).append(values.length <= i || values[i] == null ? "?" : values[i].toString());
			if (deliminator.length() == 0) deliminator = ", ";
		}
		sql.append(") VALUES (").append(valuesSql).append(")");
		if (debugEnabled) logSql.debug("INSERT SQL for VirtualDb = "+vdb+"; VirtualTable = "+table+": "+sql.toString());
		return sql.toString();
	}
	
	static String buildInsertSelectSql(VirtualDb vdb, VirtualTable table, VirtualColumn[] columns, String selectSql)
	{
		Dialect dialect = HibernateUtil.getDialect();
		char oq = dialect.openQuote();
		char cq = dialect.closeQuote();
		
		StringBuilder sql = new StringBuilder();
		sql.append("INSERT INTO ");
		sql.append(table.getQualifiedTableName(vdb));	
		sql.append(" (");
		String deliminator = "";
		for (int i = 0; i < columns.length; ++i)
		{
			sql.append(deliminator).append(oq).append(columns[i].getColumn()).append(cq);
			if (deliminator.length() == 0) deliminator = ", ";
		}
		sql.append(") ").append(selectSql);
		if (debugEnabled) logSql.debug("INSERT / SELECT SQL for VirtualDb = "+vdb+"; VirtualTable = "+table+": "+sql.toString());
		return sql.toString();
	}
	
	/**
		Build insert SQL string and put it in the SQL map.
		Also sets the property order for the class in the order map.

		@param entityName The name of the class to generate the SQL for,
			the name is usually the same as the class name, but doesn't have to be
		@param dialect The database dialect in use
	*/
	@SuppressWarnings("unchecked")
	private static void buildInsert(String entityName, Dialect dialect)
	{
		PersistentClass pClass = HibernateUtil.getClassMapping(entityName);
		assert pClass != null : "entityName '"+entityName+"' is not mapped (pClass == null)";
		char oq = dialect.openQuote();
		char cq = dialect.closeQuote();
		Table table = pClass.getTable();
		
		StringBuilder sql = new StringBuilder();
		sql.append("INSERT INTO ").append(oq).append(table.getName()).append(cq).append(" (");
		StringBuilder values = new StringBuilder("VALUES (");
		
		if (dialect.getNativeIdentifierGeneratorClass() == org.hibernate.id.SequenceGenerator.class)
		{
			// For sequence generators we must use: 
			// INSERT INTO ... (id, ...) VALUES (nextval('sequenceName'), ...)
			
			// The sequence name is found in the pClass.identifierProperty.identifierGeneratorProperties
			org.hibernate.mapping.SimpleValue idValue = (org.hibernate.mapping.SimpleValue)pClass.getIdentifierProperty().getValue();
			String sequenceName = idValue.getIdentifierGeneratorProperties().getProperty(org.hibernate.id.SequenceGenerator.SEQUENCE, "hibernate_sequence");
			sql.append(oq+"id"+cq+", ");
			values.append(dialect.getSelectSequenceNextValString(sequenceName)+", ");
		}
		
		String deliminator = "";

		List<BatchedPropertyInfo> propertyOrder = new LinkedList<BatchedPropertyInfo>();
		Iterator<Property> it = (Iterator<Property>)pClass.getPropertyIterator();
		while (it.hasNext())
		{
			Property prop = it.next();
			if (prop.isInsertable())
			{
				Iterator<Column> colIt = (Iterator<Column>)prop.getColumnIterator();
				if (colIt.hasNext())
				{
					// We assume that each propety only has one column
					Column col = colIt.next();
					sql.append(deliminator).append(oq).append(col.getName()).append(cq);
					values.append(deliminator).append("?");
					if (deliminator.length() == 0) deliminator = ", ";
					BatchedPropertyInfo bpi = 
						new BatchedPropertyInfo(prop.getName(), col.getSqlTypeCode(HibernateUtil.getMapping()));
					propertyOrder.add(bpi);
				}
			}
		}
		propertyInsertOrders.put(entityName, propertyOrder);
		sql.append(") ").append(values).append(")");
		if (debugEnabled) logSql.debug("INSERT SQL for entity = "+entityName+": "+sql.toString());
		sqlInserts.put(entityName, sql.toString());
	}
	
	/**
		Build update SQL string and put it in the SQL map.
		Also sets the property order for the class in the order map.

		@param entityName The name of the class to generate the SQL for,
			the name is usually the same as the class name, but doesn't have to be
		@param dialect The database dialect in use
	*/
	@SuppressWarnings("unchecked")
	private static void buildUpdate(String entityName, Dialect dialect)
	{
		PersistentClass pClass = HibernateUtil.getClassMapping(entityName);
		assert pClass != null : "entityName '"+entityName+"' is not mapped (pClass == null)";
		char oq = dialect.openQuote();
		char cq = dialect.closeQuote();
		Table table = pClass.getTable();
		
		StringBuilder sql = new StringBuilder();
		sql.append("UPDATE ").append(oq).append(table.getName()).append(cq).append(" SET ");
		String deliminator = "";

		List<BatchedPropertyInfo> propertyOrder = new LinkedList<BatchedPropertyInfo>();
		Iterator<Property> it = (Iterator<Property>)pClass.getPropertyIterator();
		while (it.hasNext())
		{
			Property prop = it.next();
			if (prop.isUpdateable())
			{
				Iterator<Column> colIt = (Iterator<Column>)prop.getColumnIterator();
				if (colIt.hasNext())
				{
					// We assume that each propety only has one column
					Column col = colIt.next();
					propertyOrder.add(new BatchedPropertyInfo(prop.getName(), col.getSqlTypeCode(HibernateUtil.getMapping())));
					sql.append(deliminator).append(oq).append(col.getName()).append(cq).append(" = ?");
					if (deliminator.length() == 0) deliminator = ", ";
				}
			}
		}
		sql.append(" WHERE ").append(oq).append(pClass.getIdentifierProperty().getName()).append(cq).append(" = ?");
		propertyOrder.add(new BatchedPropertyInfo(pClass.getIdentifierProperty().getName(), Types.INTEGER));
		sql.append(" AND ").append(oq).append("version").append(cq).append(" = ?");
		propertyOrder.add(new BatchedPropertyInfo("oldversion", Types.INTEGER));
		propertyUpdateOrders.put(entityName, propertyOrder);
		if (debugEnabled) logSql.debug("UPDATE SQL for entity = "+entityName+": "+sql.toString());
		sqlUpdates.put(entityName, sql.toString());
	}
	
	/**
		Build delete SQL string and put it in the SQL map.
		Also sets the property order for the class in the order map.

		@param entityName The name of the class to generate the SQL for,
			the name is usually the same as the class name, but doesn't have to be
		@param dialect The database dialect in use
	*/
	private static void buildDelete(String entityName, Dialect dialect)
	{
		PersistentClass pClass = HibernateUtil.getClassMapping(entityName);
		assert pClass != null : "entityName '"+entityName+"' is not mapped (pClass == null)";
		char oq = dialect.openQuote();
		char cq = dialect.closeQuote();
		Table table = pClass.getTable();
	
		StringBuilder sql = new StringBuilder();
		sql.append("DELETE FROM ").append(oq).append(table.getName()).append(cq);
		sql.append(" WHERE ").append(oq).append(pClass.getIdentifierProperty().getName()).append(cq).append(" = ?");
		List<BatchedPropertyInfo> propertyOrder = new LinkedList<BatchedPropertyInfo>();
		propertyOrder.add(new BatchedPropertyInfo(pClass.getIdentifierProperty().getName(), java.sql.Types.INTEGER));
		propertyDeleteOrders.put(entityName, propertyOrder);
		if (debugEnabled) logSql.debug("DELETE SQL for entity = "+entityName+": "+sql.toString());
		sqlDeletes.put(entityName, sql.toString());
	}
}
