/**
	$Id$

	Copyright (C) 2014 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Collections;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.core.data.ItemKeyData;
import net.sf.basedb.core.data.ProjectKeyData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.dbengine.DbEngine;
import net.sf.basedb.core.hibernate.DbIndexWork;
import net.sf.basedb.core.hibernate.GetCurrentCatalogWork;
import net.sf.basedb.core.hibernate.SchemaExistsWork;
import net.sf.basedb.core.hibernate.SchemaGenerator;
import net.sf.basedb.core.hibernate.SessionFactoryConfiguration;

import org.hibernate.FlushMode;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.SessionBuilder;
import org.hibernate.SessionFactory;
import org.hibernate.Transaction;
import org.hibernate.boot.registry.StandardServiceRegistryBuilder;
import org.hibernate.cfg.Configuration;
import org.hibernate.dialect.Dialect;
import org.hibernate.service.ServiceRegistry;

/**
	A special session control implementation used for services. It provides extra
	functionality that an extension can use to create database objects in a 
	separate database schema/catalog for their own use. Note that the extra
	API in this class depends on Hibernate API and may change without 
	notification in case the Hibernate API is changed (though we will try to
	give advanced notice).
	
	@author nicklas
	@since 3.4
*/
public class ServiceSessionControl
	extends SessionControl
{
	
	private static final org.slf4j.Logger log =
		org.slf4j.LoggerFactory.getLogger(ServiceSessionControl.class);
	
	ServiceSessionControl() 
	{
		super(null, "net.sf.basedb.clients.internal", false, Application.getHostName());
		login(new ServicesKeyring());
	}
	
	/**
		Create a session factory for the given configuration. This method will add
		connection information from the base.config file automatically
		to the configuration object. Note that sessions can't
		
		Classes that have been defined in the configuration should be without a
		schema definition and will then be put into the specified schema automatically.
		
		If the given schema and tables doesn't exists they will be created automatically,
		otherwise an UPDATE action is performed to modify an existing installation.
		
		@param config A configuration instance containing information needed to build
			the session factory
		@return A session factory instance
	*/
	public SessionFactory buildSessionFactory(SessionFactoryConfiguration config)
	{
		String schema = config.getSchema();
		if (schema == null)
		{
			throw new ConfigurationException("A schema must be specified");
		}
		if (log.isDebugEnabled()) log.debug("Building session factory for schema: " + schema);
		
		// Add properties from BASE
		Configuration cfg = config.getHibernateConfig();
		cfg.setProperty("hibernate.dialect", Config.getString("db.dialect", ""));
		cfg.setProperty("hibernate.connection.driver_class", Config.getString("db.driver", ""));
		
		if (config.useSeparateConnections())
		{
			cfg.setProperty("hibernate.connection.url", Config.getString("db.url", ""));
			cfg.setProperty("hibernate.connection.username", Config.getString("db.username", ""));
			cfg.setProperty("hibernate.connection.password", Config.getString("db.password", ""));
		}

		ClassLoader old = null;
		boolean hasSwitchedClassLoader = false;
		Session session = null;
		Transaction tx = null;
		SessionFactory sf = null;
		
		try
		{
			// Switch classloader since otherwise Hibernate will not find classes
			// defined in the extension's JAR file
			if (config.getClassLoader() != null)
			{
				Thread current = Thread.currentThread();
				old = current.getContextClassLoader();
				hasSwitchedClassLoader = true;
				current.setContextClassLoader(config.getClassLoader());
			}

			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			
			// Get the current database
			String catalog = HibernateUtil.doReturningWork(session, new GetCurrentCatalogWork());
			if (log.isDebugEnabled()) log.debug("Current catalog: " + catalog);
			
			// Pre-process catalog+schema so that it works for the actual database
			// Postgres leave it as it is...
			// but MySQL don't have schemas so actual catalog is a combined of both
			DbEngine dbEngine = HibernateUtil.getDbEngine();
			Dialect dialect = HibernateUtil.getDialect();
			String actualCatalog = dbEngine.getCatalogName(catalog, schema);
			String actualSchema = dbEngine.getSchemaName(catalog, schema);

			if (log.isDebugEnabled()) log.debug("Actual catalog/schema: " + actualCatalog + "/" + actualSchema);		
			if (actualCatalog != null)
			{
				cfg.setProperty("hibernate.default_catalog", actualCatalog);
			}
			if (actualSchema != null)
			{
				cfg.setProperty("hibernate.default_schema", actualSchema);
			}
			
			// Check if the actual catalog/schema exists
			boolean schemaExists = HibernateUtil.doReturningWork(session, new SchemaExistsWork(actualCatalog, actualSchema));
			SchemaGenerator.Mode mode =  schemaExists ? 
				SchemaGenerator.Mode.UPDATE : SchemaGenerator.Mode.INSTALL;
			if (log.isDebugEnabled()) log.debug("Mode: " +mode);
			
			if (mode == SchemaGenerator.Mode.INSTALL)
			{
				String sql = dbEngine.getCreateSchemaSql(actualCatalog, actualSchema);
				if (log.isDebugEnabled()) 
				{
					log.debug("Creating schema " + actualCatalog + "/" + actualSchema + ": " + sql);
				}
				Query q = session.createSQLQuery(sql);
				q.executeUpdate();
			}
			else if (config.getAppId() != null)
			{
				// Check schema version if an update is needed
				int schemaVersion = config.getSchemaVersion();
				int installedSchemaVersion = Update.getSchemaVersion(session, config.getAppId());
				if (installedSchemaVersion == schemaVersion)
				{
					// No update needed
					mode = null;
				}
				else if (installedSchemaVersion > schemaVersion)
				{
					throw new ConfigurationException("Installed schema version for schema '" + schema + 
						"' is higher than expected: " + installedSchemaVersion + ">"+schemaVersion);
				}
				config.setInstalledSchemaVersion(installedSchemaVersion);
			}
			
			if (mode != null)
			{
				// Generates tables and most other things
				SchemaGenerator schemaGenerator = 
					new SchemaGenerator(cfg, dialect, dbEngine, mode, null);
				session.doWork(schemaGenerator);
				// Generates missing indexes
				DbIndexWork indexGenerator = new DbIndexWork(cfg, dialect, dbEngine, false, true, false, true);
				session.doWork(indexGenerator);
				// Update schema version in BASE
				if (config.getAppId() != null)
				{
					Update.setSchemaVersion(session, config.getAppId(), config.getSchemaVersion(), 0);
				}
				HibernateUtil.commit(tx);
				tx = null;
			}
			
			ServiceRegistry factoryBuilder = new StandardServiceRegistryBuilder().applySettings(cfg.getProperties()).build();
			sf = cfg.buildSessionFactory(factoryBuilder);
			if (log.isDebugEnabled())
			{
				log.debug("Created session factory for schema: " + schema);
			}
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("Could not build session factory for schema: " + schema, ex);
			throw ex;
		}
		catch (Exception ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			log.error("Could not build session factory for schema: " + schema, ex);
			throw new BaseException(ex);
		}
		finally
		{
			if (hasSwitchedClassLoader) Thread.currentThread().setContextClassLoader(old);
			if (session != null) HibernateUtil.close(session);
		}
		
		return sf;
	}
	
	/**
		Create a new Hibernate session for the given session factory that
		is piggy-backed on the given DbControl transaction. 
		
		Note that the life-cycle of the returned session is managed by the
		DbControl. There is no need to flush or close the session since this
		is done automatically and may cause exceptions if done when not needed.
		
		@param builder A session builder (eg. SessionFactory.withOptions())
		@param dc An open DbControl
	*/
	public Session newSession(SessionBuilder builder, DbControl dc)
	{
		Session s = builder.connection(HibernateUtil.getConnection(dc.getHibernateSession())).openSession();
		s.setFlushMode(FlushMode.COMMIT);
		dc.addTransactionalAction(new FlushSessionAction(s));
		return s;
	}
	
	
	/**
		Our own keyring containing the only permission: {@link Permission#ACT_AS_ANOTHER_USER}.
		All other permissions are denied.
	*/
	static class ServicesKeyring
		extends Keyring
	{
	
		private final int systemPermissions;
		private final Set<Integer> emptySet;
		private final Map<Integer, Integer> emptyMap;
	
		private ServicesKeyring()
		{
			super(0);
			this.systemPermissions = Permission.grant(Permission.ACT_AS_ANOTHER_USER);
			this.emptySet = Collections.emptySet();
			this.emptyMap = Collections.emptyMap();
		}
		/*
			From the Keyring class
			-------------------------------------------
		*/
		/**
			Never reload since everything is static.
		*/
		@Override
		boolean getReload()
		{
			return false;
		}
		/**
			Return {@link Permission#ACT_AS_ANOTHER_USER} if <code>itemType</code> = 
			{@link Item#SYSTEM}, 0 otherwise. 
		*/
		@Override
		int getRolePermissions(Item itemType)
		{
			if (itemType == Item.SYSTEM)
			{
				return systemPermissions;
			}
			else
			{
				return 0;
			}
		}
		/**
			Return 0.
		*/
		@Override
		int getSharedPermissions(ShareableData sharableData)
		{
			return 0;
		}
		/**
			Return 0.
		*/
		@Override
		int getAllPermissions(Item itemType, UserData owner, ItemKeyData itemKey, ProjectKeyData projectKey)
		{
			return 0;
		}
		/**
			Return empty set.
		*/
		@Override
		Set<Integer> getRoles()
		{
			return emptySet;
		}
		/**
			Return empty set.
		*/
		Set<Integer> getGroups()
		{
			return emptySet;
		}
		/**
			Return empty set.
		*/
		@Override
		Set<Integer> getUsers()
		{
			return emptySet;
		}
		/**
			Return empty map.
		*/
		@Override
		Map<Integer, Integer> getProjects()
		{
			return emptyMap;
		}
		/**
			Return empty set.
		*/
		Set<Integer> getItemKeys()
		{
			return emptySet;
		}
		/**
			Return empty set.
		*/
		Set<Integer> getProjectKeys()
		{
			return emptySet;
		}
		// -------------------------------------------
	}

	/**
		Flushes the given session before the transaction is committed.
	*/
	static class FlushSessionAction
		implements TransactionalAction
	{
		
		private final Session session;
		
		FlushSessionAction(Session session)
		{
			this.session = session;
		}

		@Override
		public void onBeforeCommit() 
		{
			session.flush();
		}

		@Override
		public void onRollback() 
		{
			HibernateUtil.close(session);
		}
		
		@Override
		public void onAfterCommit() 
		{
			HibernateUtil.close(session);
		}
		
		
	}
	
}