/*
	$Id: Tag.java 6881 2015-04-21 07:59:42Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Date;
import java.util.Set;

import net.sf.basedb.core.data.TagData;

/**
	This class represents a tag. Tags are used on extracts so that they
	can be tracked on physical bioassays.

	@author Nicklas
	@since 3.0
	@base.modified $Date: 2015-04-21 09:59:42 +0200 (ti, 21 apr 2015) $
*/
public class Tag
	extends AnnotatedItem
	implements Registered, Subtypable
{
	/**
		The type of item represented by this class.
		@see Item#TAG
		@see #getType()
	*/
	public static final Item TYPE = Item.TAG;
	
	/**
		The ID for the LABEL tag type.
	*/
	public static final String LABEL = "net.sf.basedb.core.TagType.LABEL";

	/**
		The ID for the BARCODE tag type.
	*/
	public static final String BARCODE = "net.sf.basedb.core.TagType.BARCODE";


	/**
		Create a new <code>Tag</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@return The new <code>Label</code> item
		@throws BaseException If there is an error
	*/
	public static Tag getNew(DbControl dc)
		throws BaseException
	{
		Tag t = dc.newItem(Tag.class);
		t.setName("New tag");
		t.getData().setEntryDate(new Date());
		return t;
	}

	/**
		Get a <code>Tag</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Tag</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Tag getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Tag t = dc.loadItem(Tag.class, id);
		if (t == null) throw new ItemNotFoundException("Tag[id="+id+"]");
		return t;
	}

	/**
		Get a query that returns tags.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Tag> getQuery()
	{
		return new ItemQuery<Tag>(Tag.class);
	}

	Tag(TagData data)
	{
		super(data);
	}

	@Override
	TagData getData()
	{
		return (TagData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	// -------------------------------------------
	/*
		From the Annotatable interface
		-------------------------------------------
	*/
	/**
		@return Always null
		@since 3.3
	*/
	@Override
	public Set<Annotatable> getAnnotatableParents()
	{
		return null;
	}
	// -------------------------------------------
	/*
		From the Subtypable interface
		-----------------------------
	*/
	@Override
	public ItemSubtype getItemSubtype()
	{
		return getDbControl().getItem(ItemSubtype.class, getData().getItemSubtype());
	}
	@Override
	public void setItemSubtype(ItemSubtype subtype)
	{
		checkPermission(Permission.WRITE);
		if (subtype != null)
		{
			subtype.setOnItem(this);
		}
		else
		{
			getData().setItemSubtype(null);
		}
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>Some {@link Extract}:s are marked with this tag
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_EXTRACTS_FOR_TAG", "count(*)");
		/*
			SELECT {1}
			FROM ExtractData ext
			WHERE ext.tag = :tag
		*/
		query.setEntity("tag", this.getData());
		boolean used = HibernateUtil.loadData(Long.class, query) > 0;
		return used || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link Extract}:s marked with this tag
		</ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		
		// Labeled extracts
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_EXTRACTS_FOR_TAG", "ext.id");
		/*
			SELECT {1}
			FROM ExtractData ext
			WHERE ext.tag = :tag
		*/
		query.setEntity("tag", this.getData());
		addUsingItems(using, Item.EXTRACT, query);
		return using;
	}
	// -------------------------------------------


}
