/**
	$Id $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.encode;

/**
	Symmetrical encoder/decoder implementation that replaces newline, carriage return, 
	tab and backslash with \n, \r, \t and \\. This encoder is suitable for use with
	tab-separated text files.
	
	@author Nicklas
	@since 2.15
*/
public class TabCrLfEncoderDecoder 
	implements EncoderDecoder 
{

	private final boolean nullIsEmptyString;
	
	/**
		Creates a new encoder/decoder.
		@param nullIsEmptyString TRUE to encode null as empty string,
			FALSE to not encode null. NOTE! If TRUE, this encoder is no
			longer symmetrical
	*/
	public TabCrLfEncoderDecoder(boolean nullIsEmptyString)
	{
		this.nullIsEmptyString = nullIsEmptyString;
	}
	
	/*
		From the EncoderDecoder interface
		---------------------------------
	*/
	@Override
	public boolean isSymmetrical() 
	{
		return !nullIsEmptyString;
	}

	@Override
	public String decode(String s) 
	{
		if (s == null) return null;
		if (s.length() == 0) return nullIsEmptyString ? null : s;
		boolean neededDecode = false;
		StringBuilder sb = new StringBuilder(s.length());
		for (int i = 0; i < s.length(); ++i)
		{
			char c = s.charAt(i);
			if (c == '\\' && i < s.length() - 1)
			{
				++i;
				c = s.charAt(i);
				if (c == 't')
				{
					sb.append("\t");
					neededDecode = true;
				}
				else if (c == 'n')
				{
					sb.append("\n");
					neededDecode = true;
				}
				else if (c == 'r')
				{
					sb.append("\r");
					neededDecode = true;
				}
				else if (c == '\\')
				{
					sb.append("\\");
					neededDecode = true;
				}
				else
				{
					// Hmm... \ followed by something else than n, r, t, or \... 
					// pass through unmodified since it is better to be nice...
					sb.append("\\").append(c);
				}
			}
			else
			{
				sb.append(c);
			}
		}
		return neededDecode ? sb.toString() : s;
	}

	@Override
	public String encode(String s) 
	{
		if (s == null) return nullIsEmptyString ? "" : null;
		boolean neededEncode = false;
		StringBuilder sb = new StringBuilder(s.length());
		for (int i = 0; i < s.length(); ++i)
		{
			char c = s.charAt(i);
			if (c == '\t')
			{
				sb.append("\\t");
				neededEncode = true;
			}
			else if (c == '\n')
			{
				sb.append("\\n");
				neededEncode = true;
			}
			else if (c == '\r')
			{
				sb.append("\\r");
				neededEncode = true;
			}
			else if (c == '\\')
			{
				sb.append("\\\\");
				neededEncode = true;
			}
			else
			{
				sb.append(c);
			}
		}
		return neededEncode ? sb.toString() : s;
	}
	// -------------------------------------------

}
