/**
	$Id: TableWriter.java 5193 2009-11-27 13:06:17Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.export;

import java.io.PrintWriter;
import java.io.Writer;

import net.sf.basedb.util.encode.EncoderDecoder;
import net.sf.basedb.util.encode.IdentityEncoderDecoder;

/**
	Wraps a writer stream and provides methods for easier writing
	of tabular data. The writer can be configured with a 
	{@link #setDataSeparator(String)} (default is 'tab'),
	{@link #setNullValue(String)} (default is an empty string)
	and an {@link #setEncoder(EncoderDecoder)} (no default). 
	Use {@link #tablePrintData(Object...)} to print a singe line 
	with data. Other {@link PrintWriter} methods are not affected 
	by those settings.

	@author Nicklas
	@version 2.12
	@base.modified $Date: 2009-11-27 14:06:17 +0100 (fr, 27 nov 2009) $
*/
public class TableWriter
	extends PrintWriter
{
	
	private String dataSeparator = "\t";
	private String nullValue = "";
	private EncoderDecoder encoder;

	/**
		Create a new table writer that is writing it's output to the
		given writer.
		@param out The writer to write the data to
	*/
	public TableWriter(Writer out)
	{
		super(out);
	}

	/**
		The data separator string is written between
		each data column in the output. The default value is 'tab'.
		@return The current value of the data separator string
	*/
	public String getDataSeparator()
	{
		return dataSeparator;
	}
	
	/**
		Change the data separator string. Note! If this value is modified
		then a different encoder is probably also needed. See
		{@link #setEncoder(EncoderDecoder)}.
		@param dataSeparator The new data separator
	*/
	public void setDataSeparator(String dataSeparator)
	{
		this.dataSeparator = dataSeparator;
	}

	/**
		The null value string is written to the output instead of
		null data values. The default value is an empty string.
		@return The current value of the null value string
	*/
	public String getNullValue()
	{
		return nullValue;
	}
	
	/**
		Change the null value string.
		@param nullValue The new null value string
	*/
	public void setNullValue(String nullValue)
	{
		this.nullValue = nullValue;
	}

	/**
		Get the current default encoder for this writer.
		@return The current encoder or null
		@since 2.15
	*/
	public EncoderDecoder getEncoder()
	{
		return encoder;
	}
	
	/**
		Change the encoder.
		@param encoder The new encoder or null to not encode
		@since 2.15
	*/
	public void setEncoder(EncoderDecoder encoder)
	{
		this.encoder = encoder;
	}

	/**
		Encode the given string with the default encoder.
		@param s The string to encode
		@return The encoded string, or the argument if no default encoder
			has been specified
		@since 2.15
	*/
	public String tableEncode(String s)
	{
		EncoderDecoder e = getEncoder();
		return e == null ? s : e.encode(s);
	}

	/**
		Print a data line to the parent writer using the
		default encoder.

		@param data The values to write
	*/
	public void tablePrintData(Object... data)
	{
		tablePrintData(getEncoder(), data);
	}
	
	/**
		Print a data line to the parent writer. Each value is separated by the
		data separator string ('tab' by default). Null values are replaced 
		with the null value string (empty string by default). Non-null objects
		are converted to strings values by calling their toString() method and,
		if an encoder has been specified, then encoded with {@link 
		EncoderDecoder#encode(String)}.
	
		@param data The values to write
		@param encoder The encoder to use, or null to not use any encoder
		@since 2.15
	*/
	public void tablePrintData(EncoderDecoder encoder, Object... data)
	{
		boolean first = true;
		String ds = getDataSeparator();
		String nv = getNullValue();
		EncoderDecoder e = getEncoder();
		if (e == null) e = new IdentityEncoderDecoder();
		for (Object d : data)
		{
			if (!first) print(ds);
			print(d == null ? nv : e.encode(d.toString()));
			first = false;
		}
		print("\n");
	}
	
}
