/**
	$Id: BaseFileInfo.java 5689 2011-08-11 09:09:06Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.importer.spotdata;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DatabaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicQuery;
import net.sf.basedb.core.DynamicResultIterator;
import net.sf.basedb.core.ExtraValueType;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.util.Diff3;
import net.sf.basedb.util.importer.FileWrapper;

/**
	Collects information about a BASEfile as it gets parsed
	by a {@link BaseFileImporter}. Information in one section is
	often needed when parsing other sections or after the parsing 
	has been completed.

	@author Nicklas
	@version 2.14
	@base.modified $Date: 2011-08-11 11:09:06 +0200 (to, 11 aug 2011) $
*/
public class BaseFileInfo
{
	private final FileWrapper file;
	
	// Information about child assays
	private boolean assaysHaveParentAssaysMapping;
	private Map<Integer, ChildBioAssay> childAssays;
	
	// Information about parent bioassay set
	private boolean parentHasNullReporter;
	private boolean parentHasZeroReporter;
	// Position/reporter mapping: key = position, value=reporter id
	private Map<Integer, Integer> parentReporterPositions;
	private Map<Short, List<Short>> parentChildColumnMapping;
	
	// Information about spot sections
	private Map<Integer, SpotSectionInfo> spotSectionInfo;
	private Map<String, ExtraValueType> extraFloats;
	
	// Information about child spotdata
	private boolean childHasDifferentReporterPositionMapping;
	// Position/reporter mapping: key = position, value=reporter id
	private Map<Integer, Integer> childReporterPositions;
	
	
	/**
		The source file which should be a BASEfile.
		@since 2.16
	*/
	public BaseFileInfo(FileWrapper file)
	{
		this.file = file;
	}
	
	/**
		Get the name of the file.
	*/
	public String getName()
	{
		return file.getName();
	}
	
	/**
		Get the character set used by the file, or null if not known.
	*/
	public String getCharacterSet()
	{
		return file.getCharacterSet();
	}
	
	/**
		Get the file size.
	 */
	public long getSize()
	{
		return file.getSize();
	}
	
	/**
		Set a flag that indicates if the BASEfile includes a mapping
		from child to parent assays.
	*/
	public void setAssaysHaveParentAssaysMapping(boolean haveMapping)
	{
		this.assaysHaveParentAssaysMapping = haveMapping;
	}

	/**
		A flag that indicates if the BASEfile includes a mapping
		from child to parent assays.
	*/
	public boolean getAssaysHaveParentAssaysMapping()
	{
		return assaysHaveParentAssaysMapping;
	}

	/**
		Add information about a child assay.
		@param child Information about the child assay
		@return TRUE if the child was added, FALSE if a child with 
			the same ID already exists
	*/
	public boolean addChildAssay(ChildBioAssay child)
	{
		if (childAssays == null) childAssays = new HashMap<Integer, ChildBioAssay>();
		if (childAssays.containsKey(child.getFileId())) return false;
		childAssays.put(child.getFileId(), child);
		return true;
	}
	
	/**
		Get information about all child bioassays referenced in the file.
	*/
	public Collection<ChildBioAssay> getChildAssays()
	{
		return childAssays == null ? null : childAssays.values();
	}
	
	/**
	 	Get the child bioassay with the given id (from the BASEfile).
		@param id The ID in the BASEfile
		@return A ChildBioAssay or null
	*/
	public ChildBioAssay getChildAssay(int id)
	{
		return childAssays == null ? null : childAssays.get(id);
	}
	
	
	/**
		Checks that the bioassay with the given id exists in the database. 
		If not, an exception is thrown.
		@param dc The DbControl to use
		@param id The id of the bioassay
		@param line The current line
		@throws ItemNotFoundException If the bioassay doesn't exists in the
			database
	*/
	public void checkBioAssay(DbControl dc, int id, int line)
	{
		try
		{
			BioAssay.getById(dc, id);
		}
		catch (ItemNotFoundException ex)
		{
			throw new ItemNotFoundException("Can't find bioassay '" + id +
				"' in section 'assays' at line " + line + " in file '" + getName() + "'");
		}
	}
	
	/**
		Set a flag indicating that at least one position in the parent bioassay 
		set has a null reporter.
		@param parentHasNullReporter TRUE if there is a position with a null reporter,
			FALSE otherwise
		@see #loadParentReporterPositions(DbControl, BioAssaySet)
	*/
	public void setParentHasNullReporter(boolean parentHasNullReporter)
	{
		this.parentHasNullReporter = parentHasNullReporter;
	}
	
	/**
		A flag that indicates if at least one position in the parent bioassay
		set has a null reporter or not. We need to know this because some BASE 1
		plug-ins convert a null reporter to 0 and in most cases it is safe
		to convert it back, but only if the parent had a null reporter but not
		a '0' reporter to begin with.
		@see #getParentHasZeroReporter()
	*/
	public boolean getParentHasNullReporter()
	{
		return parentHasNullReporter;
	}
	
	/**
		Set a flag indicating that at least one position in the parent bioassay
		set has a reporter with ID=0.
	*/
	public void setParentHasZeroReporter(boolean parentHasZeroReporter)
	{
		this.parentHasZeroReporter = parentHasZeroReporter;
	}
	
	/**
		A flag that indicates if at least one position in the parent bioassay
		set has a reporter with ID=0.
		@see #getParentHasNullReporter()
	*/
	public boolean getParentHasZeroReporter()
	{
		return parentHasZeroReporter;
	}
	
	/**
		Load position/reporter mapping from the parent bioassay set. This
		will also set the {@link #getParentHasNullReporter()}/{@link #getParentHasZeroReporter()}
		flags.
		
		@param dc A DbControl to use
		@param parent The parent bioassay set
	*/
	public void loadParentReporterPositions(DbControl dc, BioAssaySet parent)
	{
		if (parentReporterPositions != null) return;
		
		parentReporterPositions = new HashMap<Integer, Integer>();
		DynamicQuery positionQuery = parent.getPositionData();
		positionQuery.select(Dynamic.select(VirtualColumn.POSITION));
		positionQuery.select(Dynamic.select(VirtualColumn.REPORTER_ID));
		DynamicResultIterator positionQueryIterator = positionQuery.iterate(dc);
		try
		{
			while (positionQueryIterator.hasNext())
			{
				SqlResult result = positionQueryIterator.next();
				Integer reporterId = (Integer)result.getObject(2);
				if (reporterId == null) 
				{
					setParentHasNullReporter(true);
				}
				else if (reporterId.intValue() == 0)
				{
					setParentHasZeroReporter(true);
				}
				parentReporterPositions.put(result.getInt(1), reporterId);
			}
		}
		catch (SQLException ex)
		{
			throw new DatabaseException(ex);
		}
	}
	
	/**
		Set a flag that indicates if the child data has a different position/reporter
		mapping than the parent bioassay set.
		@see #addChildReporter(Integer, Integer)
	*/
	public void setChildHasDifferentReporterPositionMapping(boolean flag)
	{
		this.childHasDifferentReporterPositionMapping = flag;
	}
	
	/**
		A flag that indicates if the child data has a different position/reporter
		mapping than the parent bioassay set.
	*/
	public boolean getChildHasDifferentReporterPositionMapping()
	{
		return childHasDifferentReporterPositionMapping;
	}
	
	/**
		Add a child position/reporter mapping from the parsed BASEfile.
		The added mapping will be checked against the parent position/reporter
		mapping and the {@link #getChildHasDifferentReporterPositionMapping()}
		flag is set if the mapping has changed.
		<p>
		If a mapping for the given position already exists, a check is made
		if the mapping is identical or different. If the mapping is different
		the reporter id of the already existing mapping is returned. It is
		up to the called to take any appropriate action in this case (usually;
		throw an exception). If the mapping doesn't exist or if the mapping
		is identical the 'reporterId' value is returned. Thus, it is easy to
		check for success by simply comparing:
<pre class="code">		
if (reporterId != addChildReporter(position, reporterId))
{
   // ... throw exception
}
</pre>		
		@param position The position value
		@param reporterId The ID of the reporter
		@return The reporterId (same instance) if the mapping is new or
			if an identical mapping already exists. The reporer ID of an
			existing mapping if it is different. 
	*/
	public Integer addChildReporter(Integer position, Integer reporterId)
	{
	
		// First we check if the position reporter is the same as for the parent
		if (!childHasDifferentReporterPositionMapping)
		{			
			Integer parentReporterId = parentReporterPositions.get(position);
			if (!Diff3.isEqualOrNull(reporterId, parentReporterId))
			{
				setChildHasDifferentReporterPositionMapping(true);
				parentReporterPositions.clear();
			}
		}
		
		Integer returnValue = reporterId;
		
		// Then, we check if the position has already been assigned a 
		// reporter or not
		if (childReporterPositions == null) childReporterPositions = new HashMap<Integer, Integer>();
		if (!childReporterPositions.containsKey(position))
		{
			// ... new position
			childReporterPositions.put(position, reporterId);
		}
		else
		{
			// Error if same position has already been assigned a different reporter
			Integer registeredReporterId = childReporterPositions.get(position);
			if (!Diff3.isEqualOrNull(reporterId, registeredReporterId))
			{
				returnValue = registeredReporterId;
			}
		}
		return returnValue;
	}

	/**
		Get a map with all added child position/reporter mappings. 
		The key is the position value and the value is the reporter id.
	*/
	public Map<Integer, Integer> getChildReporterPositions()
	{
		return childReporterPositions;
	}
	
	/**
		Maps a parent data cube column to a child cube column.
		Note! A parent may map to multiple childs, and a child may 
		have multiple parents.
		
		@param parentColumn The parent data cube column
		@param childColumn The child data cube column
	*/
	public void mapParentChildColumns(short parentColumn, short childColumn)
	{
		if (parentChildColumnMapping == null) 
		{
			parentChildColumnMapping = new HashMap<Short, List<Short>>();
		}
		List<Short> childColumns = parentChildColumnMapping.get(parentColumn);
		if (childColumns == null) childColumns = new ArrayList<Short>();
		childColumns.add(childColumn);
		parentChildColumnMapping.put(parentColumn, childColumns);
	}
	
	/**
		Get all child columns that have been mapped to the specified parent
		column.
		@return A list or null if no mapping exists
	*/
	public List<Short> getChildColumns(short parentColumn)
	{
		return parentChildColumnMapping == null ? 
			null : parentChildColumnMapping.get(parentColumn);
	}
	
	/**
		Adds information about a 'spot' section.
		@param spotSection The ordinal number of the spot section as it
			appears in the BASEfile, starting with 1
		@param info The information
	*/
	public void addSpotSectionInfo(int spotSection, SpotSectionInfo info)
	{
		if (spotSectionInfo == null) spotSectionInfo = new HashMap<Integer, SpotSectionInfo>();
		spotSectionInfo.put(spotSection, info);
	}
	
	/**
		Get information about a spot section.
		@param spotSection The ordinal number of the spot section as it
			appears in the BASEfile, starting with 1
	*/
	public SpotSectionInfo getSpotSectionInfo(int spotSection)
	{
		return spotSectionInfo.get(spotSection);
	}
	
	/**
		Add information about an extra float value that appears in the
		BASEfile. This method will check if the an {@link ExtraValueType}
		with the given external id exists. If not, a new float-type extra
		value type is created.
		@param dc A DbControl object to use for database access
		@param extraFloatId The external id of an extra value type
	*/
	public ExtraValueType addExtraFloat(DbControl dc, String extraFloatId)
	{
		if (extraFloats == null) extraFloats = new HashMap<String, ExtraValueType>();
		ExtraValueType extra = extraFloats.get(extraFloatId);
		if (extra == null)
		{
			try
			{
				extra = ExtraValueType.getByExternalId(dc, extraFloatId);
			}
			catch (ItemNotFoundException e)
			{
				extra = ExtraValueType.getNew(dc, extraFloatId, Type.FLOAT);
				extra.setName(extraFloatId);
				dc.saveItem(extra);
			}
			extraFloats.put(extraFloatId, extra);
		}
		return extra;
	}
	
	/**
		Get the extra value type that has the given external id. 
		The extra float must have been added with {@link #addExtraFloat(DbControl, String)}
		before.
		
		@param extraFloatId The external id of an extra value type
		@return The extra value type, or null
	*/
	public ExtraValueType getExtraFloat(String extraFloatId)
	{
		return extraFloats == null ? null : extraFloats.get(extraFloatId);
	}
	
	/**
		Keeps information about a new child bioassay from
		the 'assays' section or 'spot' sections.
	*/
	public static class ChildBioAssay
	{
		private int fileId;
		private String name;
		private Set<Integer> parents;
		private BioAssay child;
		
		/**
			Create a new object.
		*/
		public ChildBioAssay()
		{}
		
		/**
			Create a new object
		*/
		public ChildBioAssay(int fileId)
		{
			this.fileId = fileId;
		}
		
		/**
			Get the id of the child bioassay as it appears in the 
			BASEfile. This is either the same as the id of a parent bioassay
			or an internal id only.
		*/
		public int getFileId()
		{
			return fileId;
		}
		
		/**
			Set the id of the child bioassay as it appears in the BASEfile.
		*/
		public void setFileId(int fileId)
		{
			this.fileId = fileId;
		}
		
		/**
			Get the name of the child bioassay. May return null if the
			BASEfile doesn't specify a name in which case the child will
			get the same name as it's parent.
		*/
		public String getName()
		{
			return name;
		}
		
		/**
			Set the name of the child bioassay. If null, the child
			will be given the same name as it's parent.
		*/
		public void setName(String name)
		{
			this.name = name;
		}
		
		/**
			Get the parent assays of this child. If no parents has been 
			specified, the parent assay is the one that has the same id
			as the file id.
		*/
		public Set<Integer> getParents()
		{
			return parents;
		}
		
		/**
			Set the parents for this child.
		*/
		public void setParents(Set<Integer> parents)
		{
			this.parents = parents;
		}
		
		/**
			Get the real child bioassay object once it has been created
			and scheduled for saving into the database.
			@see BaseFileImporter#createChildBioAssays(DbControl, BioAssaySet, BaseFileInfo, boolean)
		*/
		public BioAssay getChild()
		{
			return child;
		}
		
		/**
			Set the real child bioassay object.
		*/
		public void setChild(BioAssay child)
		{
			this.child = child;
		}
		
	}

	/**
		Keeps information about headers and more in a 'spot' section.
	*/
	public static class SpotSectionInfo
	{
		private int positionIndex;
		private int firstAssayFieldIndex;
		private List<Integer> assays;
		private List<String> assayFields;
		private List<ExtraFloatParser> extraFloats;
	
		private SpotIntensityParser spiParser;
	
		/**
			Creates a new object.
		*/
		public SpotSectionInfo()
		{}
		
		/**
			The index of the 'position' data column.
		*/
		public int getPositionIndex()
		{
			return positionIndex;
		}
		
		/**
			Set the column index of the 'position' data column.
		*/
		public void setPositionIndex(int positionIndex)
		{
			this.positionIndex = positionIndex;
		}
		
		/**
			The index of the first data column that contains assay data.
		*/
		public int getFirstAssayFieldIndex()
		{
			return firstAssayFieldIndex;
		}
		
		/**
			Set the index of the first data column that contains assay data.
		*/
		public void setFirstAssayFieldIndex(int firstAssayFieldIndex)
		{
			this.firstAssayFieldIndex = firstAssayFieldIndex;
		}

		/**
			Get a list with all assay data fields, in the order they
			appear in the file.
		*/
		public List<String> getAssayFields()
		{
			return assayFields;
		}
		
		/**
			Set the list of all assay data fields, in the order
			the appear in the file.
		*/
		public void setAssayFields(List<String> assayFields)
		{
			this.assayFields = assayFields;
		}

		/**
			Get a list with the ID:s of all assays in the section,
			in the order they appear in the data columns.
		*/
		public List<Integer> getAssays()
		{
			return assays;
		}
		
		/**
			Set a list with the ID:s of all assays, in the
			order the appearu in the data columns.
		*/
		public void setAssays(List<Integer> assays)
		{
			this.assays = assays;
		}
		
		/**
			Get the spot intensity parser that should be used for
			parsing spot intensities in this section.
		*/
		public SpotIntensityParser getSpotIntensityParser()
		{
			return spiParser;
		}
		
		/**
			Set the spot intensity parser for the section.
		*/
		public void setSpotIntensityParser(SpotIntensityParser spiParser)
		{
			this.spiParser = spiParser;
		}

		/**
			Get the list of extra value parsers that exists for
			this section.
		*/
		public List<ExtraFloatParser> getExtraFloatParsers()
		{
			return extraFloats;
		}
	
		/**
			Add an extra value parser to this section.
		*/
		public void addExtraFloatParser(ExtraFloatParser extraFloatParser)
		{
			if (extraFloats == null) extraFloats = new LinkedList<ExtraFloatParser>();
			extraFloats.add(extraFloatParser);
		}
	
	}
}
