/**
	$Id: OverviewUtil.java 6960 2015-10-01 10:59:58Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview;

import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import java.util.HashSet;
import java.util.Set;

import org.apache.commons.collections.IteratorUtils;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.Application;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Presets;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.Subtypable;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.AnnotationUtil;
import net.sf.basedb.util.AnnotationUtil.Cache;
import net.sf.basedb.util.extensions.ClientContext;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.manager.Settings;
import net.sf.basedb.util.overview.extensions.ValidationRuleAction;
import net.sf.basedb.util.overview.filter.BasicItemFilter;
import net.sf.basedb.util.overview.Node;

/**
	Utility class for the overview generation functionality.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2015-10-01 12:59:58 +0200 (to, 01 okt 2015) $
*/
public final class OverviewUtil
{

	private static Map<String, List<ValidationRuleAction>> staticRules;

	/**
		@return Same as {@link #getStaticRules()}
		@deprecated In 3.2, use {@link #getStaticRules()} or {@link #getAllRules(DbControl, GenericOverview)} instead
	*/
	@Deprecated
	public static Map<String, List<ValidationRuleAction>> getValidators()
	{
		return getStaticRules();
	}

	/**
		Get all static (built-in) validation rules grouped by relationship. The key of the map
		is a title that groups related validation rules, for example:
		"Project default", "Missing items", etc.
		@return A map with a string key and list of validation rules 
			as value
		@since 3.2
	*/
	public static Map<String, List<ValidationRuleAction>> getStaticRules()
	{
		if (staticRules == null)
		{
			Map<String, List<ValidationRuleAction>> temp = new LinkedHashMap<String, List<ValidationRuleAction>>();
			temp.put("Project defaults", 
					Arrays.asList(new ValidationRuleAction[] 
					{
						Validator.NONDEFAULT_PLATFORM,
						Validator.NONDEFAULT_RAWDATATYPE, Validator.NONDEFAULT_ARRAYDESIGN, 
						Validator.NONDEFAULT_PROTOCOL, Validator.NONDEFAULT_HARDWARE, 
						Validator.NONDEFAULT_SOFTWARE
					})
				);
			temp.put("Missing items", 
					Arrays.asList(new ValidationRuleAction[] 
					{
						Validator.MISSING_BIOSOURCE, Validator.MISSING_SAMPLE, Validator.MISSING_EXTRACT,
						Validator.MISSING_BIOPLATE, Validator.MISSING_PHYSICALBIOASSAY, 
						Validator.MISSING_DERIVEDBIOASSAY, Validator.MISSING_ARRAYDESIGN, Validator.MISSING_PROTOCOL, 
						Validator.MISSING_HARDWARE, Validator.MISSING_SOFTWARE, Validator.MISSING_TAG,
						Validator.MISSING_ARRAYSLIDE, Validator.MISSING_ARRAYBATCH, Validator.MISSING_ANYTOANY
					})
				);
			temp.put("Subtypes", 
					Arrays.asList(new ValidationRuleAction[]
					{
						Validator.INCORRECT_BIOSOURCE_TYPE, Validator.INCORRECT_SAMPLE_TYPE,
						Validator.INCORRECT_EXTRACT_TYPE, Validator.INCORRECT_BIOMATERIAL_TYPE,
						Validator.INCORRECT_PHYSICALBIOASSAY_TYPE,
						Validator.INCORRECT_DERIVEDBIOASSAY_TYPE, Validator.INCORRECT_PROTOCOLTYPE,
						Validator.INCORRECT_HARDWARETYPE, Validator.INCORRECT_SOFTWARETYPE,
						Validator.INCORRECT_TAGTYPE
					})
				);
			temp.put("Annotations",
					Arrays.asList(new ValidationRuleAction[]
					{
						Validator.MISSING_MIAME_ANNOTATION, Validator.MISSING_FACTORVALUE, 
						Validator.MISSING_PARAMETER, Validator.ANNOTATION_IS_PARAMETER, 
						Validator.ANNOTATION_INVALID_VALUE, Validator.ANNOTATION_TOO_MANY_VALUES,
						Validator.ANNOTATION_INVALID_ITEM,
						Validator.ANNOTATION_INHERIT_FROM_NONPARENT, 
						Validator.ANNOTATION_CLONED_FROM_UNKNOWNPARENT, Validator.ANNOTATION_CLONE_OUTOFSYNC,
						Validator.ANNOTATION_INHERIT_FROM_DENIEDPARENT,
						Validator.ANNOTATION_INHERIT_MULTIPLE
					})
				);
			
			temp.put("Files", 
					Arrays.asList(new ValidationRuleAction[]
					{
						Validator.DATAFILE_INVALID, Validator.DATAFILE_NOTVALIDATED,
						Validator.MISSING_DATAFILE, Validator.DATAFILE_NOTPLATFORM,
						Validator.DATAFILE_TOOMANYFILES,
						Validator.DATAFILE_INVALID_ITEM, Validator.DENIED_DATAFILETYPE,
						Validator.DENIED_FILE
					})
				);
			
			temp.put("Denied access",
					Arrays.asList(new ValidationRuleAction[]
					{
						Validator.DENIED_BIOSOURCE, Validator.DENIED_SAMPLE, Validator.DENIED_EXTRACT, Validator.DENIED_BIOPLATE,
						Validator.DENIED_TAG, Validator.DENIED_PHYSICALBIOASSAY, Validator.DENIED_DERIVEDBIOASSAY,
						Validator.DENIED_ARRAYDESIGN, Validator.DENIED_PLATFORM, Validator.DENIED_PROTOCOL, Validator.DENIED_HARDWARE,
						Validator.DENIED_SOFTWARE, Validator.DENIED_ARRAYSLIDE, Validator.DENIED_ARRAYBATCH,
						Validator.DENIED_ANYTOANY
					})
				);
			
			temp.put("Link consistency", 
					Arrays.asList(new ValidationRuleAction[]
					{
						Validator.NONMATHCING_ARRAYDESIGN, Validator.MULTIPLE_ARRAYDESIGNS,
						Validator.CIRCULAR_REFERENCE,
						Validator.NONUNIQUE_TAG, Validator.DIFFERENT_TAG, Validator.TAGGED_PARENT,
						Validator.PARENT_EXTRACT_MISMATCH
						
					})
				);
			
			temp.put("Other",
					Arrays.asList(new ValidationRuleAction[]
					{
						Validator.NONMATCHING_SPOTCOUNT, Validator.INCORRECT_BIOASSAY_POSITION,
						Validator.NONUNIQUE_NAME_FOR_TYPE, Validator.NONUNIQUE_NAME_GLOBAL
					})
				);
			
			staticRules = Collections.unmodifiableMap(temp);
		}
		
		return staticRules;
	}

	/**
		Get all static and extension validation rules grouped by relationship. The key of the map
		is a title that groups related validation rules, for example:
		"Project default", "Missing items", etc.
		@param dc An open DbControl
		@param overview The overview tree
		@return A map with a string key and list of validation rules 
			as value
		@since 3.2
	*/
	@SuppressWarnings("unchecked")
	public static Map<String, List<ValidationRuleAction>> getAllRules(DbControl dc, GenericOverview overview)
	{
		// Get static rules
		Map<String, List<ValidationRuleAction>> allRules = new LinkedHashMap<String, List<ValidationRuleAction>>(getStaticRules());
		// Check extensions registry for more validation rules
		Registry registry = Application.getExtensionsManager().getRegistry();
		Settings settings = Application.getExtensionsManager().getSettings();
		ExtensionsInvoker<ValidationRuleAction> invoker = 
			(ExtensionsInvoker<ValidationRuleAction>)registry.useExtensions(new ClientContext(dc, overview), settings, "net.sf.basedb.util.overview.validationrule");
		
		List<ValidationRuleAction> xtRules = (List<ValidationRuleAction>)IteratorUtils.toList(invoker.iterator());
		if (xtRules != null && xtRules.size() > 0)
		{
			allRules.put("Extensions", xtRules);
		}
		return allRules;
	}

	
	/**
		Name of setting to store saved presets. DO NOT CHANGE the name since
		since it is used in the database to store the presets.
	*/
	private static final String presetsSetting = "experiment.overview.presets";
	
	/**
		Get the validation options presets for the currently logged in user.
		If changes are made to the presets call {@link #saveValidationPresets(SessionControl)}
		to make sure the changes are saved to the database.
		
		@param sc The session control for the logged in user
		@return The validation option presets
	*/
	public static Presets getValidationPresets(SessionControl sc)
	{
		Presets presets = (Presets)sc.getSessionSetting(presetsSetting);
		if (presets == null)
		{
			String xml = sc.getUserDefaultSetting(presetsSetting);
			presets = new Presets();
			if (xml != null)
			{
				presets.loadFrom(xml);
			}
			sc.setSessionSetting(presetsSetting, presets);
		}
		return presets;
	}
	
	/**
		Save the validation option presets for the logged in user. The presets
		are stored as user default settings 
		({@link SessionControl#setUserDefaultSetting(String, String)})
		with the name <code>experiment.overview.presets</code>
		@param sc The session control for the logged in user 
	*/
	public static void saveValidationPresets(SessionControl sc)
	{
		Presets presets = getValidationPresets(sc);
		if (presets != null)
		{
			sc.setUserDefaultSetting(presetsSetting, presets.asXml());
		}
	}
	
	/**
		Create a new overview for the given item and project. The created
		overview is cached and can be retrieved with {@link 
		#getCurrentOverview(SessionControl)} or {@link 
		#getNewOrCurrentOverview(DbControl, BasicItem, Project)}.
		
		@param dc A DbControl to use for database access
		@param item The item to generate the overview for
		@param project The project that should be used for validation
			of default items
		@return An overview
		@see #getNewOrCurrentOverview(DbControl, BasicItem, Project)
		@see #getCurrentOverview(SessionControl)
	*/
	public static GenericOverview getNewOverview(DbControl dc, BasicItem item, Project project)
	{
		SessionControl sc = dc.getSessionControl();
		GenericOverview overview = new GenericOverview(dc, item, project);
		sc.setSessionSetting("generic.overview", overview);
		return overview;
	}
	
	/**
		Get the current overview or create a new overview. This method will first
		check if a cached overview for the same item/project exists. If it does,
		the existing overview is returned, otherwise a new overview is created.
		
		@param dc A DbControl to use for database access
		@param item The item to generate the overview for
		@param project The project that should be used for validation
			of default items
		@return An overview
		@see #getNewOverview(DbControl, BasicItem, Project)
		@see #getCurrentOverview(SessionControl)
	*/
	public static GenericOverview getNewOrCurrentOverview(DbControl dc, BasicItem item, Project project)
	{
		GenericOverview overview = getCurrentOverview(dc.getSessionControl());
		if (overview != null)
		{
			boolean sameItem = item.equals(overview.getRootItem());
			boolean sameProject = (project==null && overview.getProject() == null) || 
				(project != null && project.equals(overview.getProject()));
			if (!sameItem || !sameProject) overview = null;
		}
		if (overview == null)
		{
			overview = getNewOverview(dc, item, project);
		}
		return overview;
	}
	
	/**
		Get the currently cached overview.
		@param sc The session control that contains the cached overview
		@return An overview, or null if no current overview exists
	*/
	public static GenericOverview getCurrentOverview(SessionControl sc)
	{
		return (GenericOverview)sc.getSessionSetting("generic.overview");
	}
	
	/**
		Find the first parent node that has a reference to the given item.
		@param node The node to start looking at
		@param item The item to look for
		@return The first parent node (including the given) that has a reference to 
			the given item, or null if no such node exists
	*/
	public static Node getCircularRef(Node node, BasicItem item)
	{
		if (item.equals(node.getItem())) return node;
		return node.getFirstParent(new BasicItemFilter(item));
	}
	
	/**
		Add a restriction to the query that limit the returned result to those items
		that has a subtype with the {@link ItemSubtype#getPushAnnotations()} flag set.
		@since 3.1
	*/
	public static void restrictToPushAnnotationSubtypes(ItemQuery<? extends Subtypable> query)
	{
		query.join(Hql.innerJoin("itemSubtype", "sbt"));
		query.restrict(Restrictions.eq(Hql.property("sbt", "pushAnnotations"), Expressions.bool(true)));
	}
	
	/**
		Get all annotation types that has been marked as "Required for MIAME".
		The annotation types found by this method will be cached in the overview
		context. Subsequent calls will return the cached annotation types.
		
		@param dc A DbControl to use for database access
		@param context The current overview context (will be used to prepare
			the query for annotation types; see {@link OverviewContext#initQuery(ItemQuery, String)}
		@param itemType An optional item type that is used to limit what
			annotation types to look for
		@return A set with the annotation types
	*/
	@SuppressWarnings("unchecked")
	public static Set<AnnotationType> getMiameAnnotationTypes(DbControl dc, OverviewContext context, Item itemType)
	{
		String cacheKey = "miame.annotationtypes";
		if (itemType != null) cacheKey += "." + itemType.name();
		Set<AnnotationType> miameAnnotationTypes = 
			(Set<AnnotationType>)context.getCachedObject(cacheKey);
		
		if (miameAnnotationTypes == null)
		{
			miameAnnotationTypes = new HashSet<AnnotationType>();
			context.setCachedObject(cacheKey, miameAnnotationTypes);
			
			// Load 'Required for MIAME' annotation types (parameter='false' must also be set)
			ItemQuery<AnnotationType> query = context.initQuery(AnnotationType.getQuery(itemType), "name");
			query.restrict(
				Restrictions.eq(
					Hql.property("requiredForMiame"), 
					Expressions.parameter("miameFlag", true, Type.BOOLEAN)
				)
			);
			query.restrict(
				Restrictions.eq(
					Hql.property("protocolParameter"), 
					Expressions.parameter("parameterFlag", false, Type.BOOLEAN)
				)
			);
			ItemResultIterator<AnnotationType> it = query.iterate(dc);
			while (it.hasNext())
			{
				miameAnnotationTypes.add(it.next());
			}
		}
		return miameAnnotationTypes;
	}
	
	/**
		Get all annotatable parent items for the given item. The parents that are found by
		this method will be cached in the current overview context. Subsequent calls will
		return the cached information.
		@param dc A DbControl to use for database access
		@param context The current overview context
		@param item The annotatable item that we should find parents for
		@return A set with the parents
	*/
	public static Set<Annotatable> getAnnotatableParents(DbControl dc, OverviewContext context, Annotatable item)
	{
		Cache cache = (Cache)context.getCachedObject("annotatable.parent.cache");
		if (cache == null)
		{
			cache = AnnotationUtil.createCache();
			context.setCachedObject("annotatable.parent.cache", cache);
		}
		return AnnotationUtil.getAllAnnotatableParentItems(dc, item, null, cache);
	}

	
}
