/*
	$Id $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.executor;

import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.ParameterValues;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;

/**
	This interface is a tagging interface that should be implemented
	by export and import plug-ins that knows how to cooperate with the external
	program executor. The plug-ins may or may not be usable as standalone plug-ins.
	The cooperation with the external program executor is designed to not interfere 
	(too much) with regular usage of the export plug-in. 
	<p>
	
	The communication between the external executor and the importer/exporter
	uses the regular plug-in API. Eg. 
	{@link InteractivePlugin#getRequestInformation(GuiContext, String)},
	{@link InteractivePlugin#configure(GuiContext, Request, Response)}
	and {@link Plugin#run(Request, Response, ProgressReporter)}.
	But we use a different set of commands than what is used in the
	normal plug-in execution sequence, so it is possible for the import/export 
	plug-ins to know if they are invoked by the executor or by the 
	regular plug-in execution sequence.
	<p>
	
	Note! The external executor doesn't check the return values of methods such
	as {@link Plugin#supportsConfigurations()}, {@link Plugin#requiresConfiguration()},
	{@link InteractivePlugin#getGuiContexts()}, etc. By implementing this interface
	a plug-in is assumed to work with the external executor. This also means that
	a plug-in may use the above mentioned methods to "hide" itself as a regular
	plug-in (eg. if the <code>getGuiContexts()</code> returns null it will never
	show up in the interface). See {@link AbstractIOSupport} which can be used as
	a base class for such plug-ins.
	<p>
	
	The external program executor has a three-step configuration sequence.
	<ol>
	<li>Select external program, exporter and importer and specify
		pre-conditions for the bioassay set such as raw data type,
		number of channels etc.
	<li>Configure the exporter and importer. This step is forwarded
		to the importer/exporter plug-ins selected in the first step
		with the command {@link #COMMAND_CONFIGURE_IO_PLUGIN}. The
		configuration parameters are proxied from the external executor
		which means that the exporter/importer has access to the
		options selected in the first step. The import/export plug-ins
		may only respond with {@link net.sf.basedb.core.plugin.Response.Status#DONE}
		or {@link net.sf.basedb.core.plugin.Response.Status#ERROR}.
		Other responses are treated as an error. Configuration settings
		that are stored by the importer and exporter are merged with the
		external executors configuration and stored in the database.
	<li>Define parameters for the external program.
	
	</ol>
	
	The job configuration sequence .... TODO
	
	
	@author nicklas
	@since 2.15
	@see ExternalProgramExecutor
	@base.modified $Date $
*/
public interface IOSupport
	extends InteractivePlugin
{

	/**
		This command is sent to import and export plug-ins in the plug-in 
		configuration sequence:
		{@link InteractivePlugin#getRequestInformation(GuiContext, String)}
		and {@link InteractivePlugin#configure(GuiContext, Request, Response)}
	*/
	public static final String COMMAND_CONFIGURE_IO_PLUGIN = "CONFIG_IO_PLUGIN";

	/**
		This command is sent to import and export plug-ins in the job 
		configuration sequence:
		{@link InteractivePlugin#getRequestInformation(GuiContext, String)}
		and {@link InteractivePlugin#configure(GuiContext, Request, Response)}
	*/
	public static final String COMMAND_CONFIGURE_IO_JOB = "CONFIG_IO_JOB";

	/**
		This command is sent to import and export plug-ins in the job 
		execution sequence: {@link Plugin#run(Request, Response, ProgressReporter)}
	*/
	public static final String COMMAND_EXECUTE_IO_JOB = "EXECUTE_IO_JOB";

	/**
		Get a short name of the plug-in. This is used by the external program 
		executor in the plug-in selection lists during the configuration
		phase.
		
		@since 2.16
	*/
	public String getShortName();
	
	/**
		Replaces the regular plug-in initialisation method. When the plug-in
		is used as part of the external executor this method is used instead
		of {@link Plugin#init(SessionControl, ParameterValues, ParameterValues)}.
		@param executor The external executor plug-in
	*/
	public void init(SessionControl sc, ParameterValues configuration, 
		ParameterValues job, ExternalProgramExecutor executor);
	
}
