/*
	$Id: TestChangeHistory.java 6875 2015-04-20 09:08:18Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

import java.util.Date;

import net.sf.basedb.core.Application;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.ChangeHistory;
import net.sf.basedb.core.Sample;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultList;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.User;
import net.sf.basedb.core.log.LogManagerFactory;
import net.sf.basedb.core.log.db.DbLogManagerFactory;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.ConsoleProgressReporter;
import net.sf.basedb.util.extensions.Extension;
import net.sf.basedb.util.extensions.ExtensionBean;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.manager.Settings;

public class TestChangeHistory
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing change history");
		write_header();
		
		ChangeHistory.deleteStrayEntries(null);
		
		// Register a new instance of the DbLogger as an extension
		// If the core logger is enabled, disable it
		Settings xtSettings = Application.getExtensionsManager().getSettings();
		Registry xtRegistry = Application.getExtensionsManager().getRegistry();
		Extension<?> coreLogger = xtRegistry.getExtension("net.sf.basedb.core.db-log-manager");
		boolean coreLoggerWasEnabled = false;
		if (coreLogger != null && xtSettings.isEnabled(coreLogger))
		{
			coreLoggerWasEnabled = true;
			xtSettings.enableExtension(coreLogger.getId(), false);
		}
		
		Extension<LogManagerFactory> logger = new ExtensionBean<LogManagerFactory>(
			"test.db-logger", "net.sf.basedb.core.log-manager", 1, null, new DbLogManagerFactory(true, true), null);
		xtRegistry.registerExtension(logger, null);
		
		//LogManagerFactory old = set_log_manager(new DbLogManagerFactory());
		Date since = new Date();
		int bioSourceId = TestBioSource.test_create("Biosource 1", false);
		int sampleId = TestSample.test_create(bioSourceId, "Sample 1", 0, false);
		test_update_sample(sampleId);
		int annotationTypeId = TestAnnotationType.test_create("change", Type.INT, 0, null, null, Item.SAMPLE, 1, null, false);
		TestAnnotation.test_annotatate(Item.SAMPLE, sampleId, annotationTypeId, 0, 1);
		TestAnnotation.test_annotatate(Item.SAMPLE, sampleId, annotationTypeId, 0, 2);
		TestAnnotation.test_remove_annotation(Item.SAMPLE, sampleId, annotationTypeId);
		int extractId = TestExtract.test_create(sampleId, "Extract 1", 0, 0, false);
		TestExtract.test_delete(extractId);
		
		// 1 CREATE event only
		test_list_by_item(Item.BIOSOURCE, bioSourceId, 1);
		
		/*
			1 CREATE
			2 UPDATE (name and externalId)
			3 ADD+UPDATE+REMOVE ('change' annotation)
			2 UPDATE (remainingQuantity when creating/deleting extract)
		*/
		test_list_by_item(Item.SAMPLE, sampleId, 8);
		/*
		 	Total change log events:
		 	4 CREATE (biosource, sample, annotationtype, extract)
		 	1 ADD (annotation)
		 	5 UPDATE (when updating the sample, it's annotation and child extract)
		 	1 REMOVE (annotation)
		 	1 DELETE (extract)
		*/
		test_list_by_user(TestUtil.getSessionControl().getLoggedInUserId(), since, 12);
		
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		// Standard test: Delete
		TestSample.test_delete(sampleId);
		TestBioSource.test_delete(bioSourceId);
		TestAnnotationType.test_delete(annotationTypeId);
		// 15 = The original 12 create+update + 3 delete
		test_delete_stray(15);
		
		xtRegistry.unregisterExtension("test.db-logger");
		
		if (coreLoggerWasEnabled)
		{
			xtSettings.enableExtension(coreLogger.getId(), true);
		}
		
		write("++Testing change history "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static void test_list_by_item(Item itemType, int itemId, int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BasicItem item = itemType.getById(dc, itemId);
			ItemQuery<ChangeHistory> q = ChangeHistory.getHistoryOf(item);
			ItemResultList<ChangeHistory> l = q.list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List history of " + item + " OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List history of " + itemType + "[id=" + itemId + "] FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list_by_user(int userId, Date since, int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			User user = User.getById(dc, userId);
			ItemQuery<ChangeHistory> q = ChangeHistory.getChangesBy(user);
			if (since != null)
			{
				q.restrict(Restrictions.gteq(Hql.property("hst", "time"), 
						Expressions.parameter("time", since)));
			}
			ItemResultList<ChangeHistory> l = q.list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List changes by user OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List changes by user FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}


	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tTime     \tType\tItem  \tChange info\tOld value\tNew value\tUser");
			write("-- \t-- \t---------\t----\t------\t-----------\t---------\t---------\t----");
		}
	}
	static void write_item(int i, ChangeHistory c)
		throws BaseException
	{
		String item = null;
		try
		{
			BasicItem b = c.getItem(null);
			item = b.toString();
		}
		catch (ItemNotFoundException ex)
		{
			item = c.getItemType() + "[id=" + c.getItemId() + "; not found]";
		}
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+c.getId()+"\t"+c.getTime()+"\t"+
			c.getChangeType() + "\t" + item + "\t" + c.getChangeInfo()+"\t"+c.getOldValue()+"\t"+c.getNewValue()+"\t"+c.getUser(null));
	}
	static void write(String message)
	{
		System.out.println(message);
	}

	static void test_delete_stray(int expectedResults)
	{
		try
		{
			ProgressReporter progress = TestUtil.getSilent() ? null : new ConsoleProgressReporter();
			int deletedLinks = ChangeHistory.deleteStrayEntries(progress);
			if (expectedResults >= 0 && expectedResults != deletedLinks)
			{
				throw new BaseException("Expected "+expectedResults+" deletions, not "+deletedLinks);
			}
			if (progress != null) progress.append("\n");
			write("--Delete stray change history OK ("+deletedLinks+")");
		}
		catch (Throwable ex)
		{
			write("--Delete stray change history FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{}
	}

	static void test_update_sample(int sampleId)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			s.setName("Updated name");
			s.setExternalId("Added external id");
			dc.commit();
			write("--Updated sample OK");
		}
		catch (Throwable ex)
		{
			write("--Updated sample FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
}

