/**
	$Id: ExportTest.java 5827 2011-10-26 10:51:52Z nicklas $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.test.performance;


import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.ExtendedProperties;
import net.sf.basedb.core.ExtendedProperty;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.RawDataProperty;
import net.sf.basedb.core.RawDataTypes;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.test.PluginUtil;
import net.sf.basedb.test.TestUtil;
import net.sf.basedb.util.Values;

/**
	Create jobs for the bioassay set exporter.

	@author nicklas
	@version 2.11
	@base.modified $Date: 2011-10-26 12:51:52 +0200 (on, 26 okt 2011) $
*/
public class ExportTest
{

	public static boolean ok = true;

	/**
		Run the tests in this class.
	*/
	public static boolean run(String user, String password, String sourcePrefix, boolean execute)
	{
		DbControl dc = null;
		TestUtil.write("++Entering export test\n");
		try
		{
			TestUtil.login(user, password);

			// Configuration options
			TestUtil.write("---------------------------------------------\n");
			TestUtil.write("Source prefix=" + sourcePrefix + "\n");
			TestUtil.write("      Execute=" + execute + "\n");
			TestUtil.write("---------------------------------------------\n");
			
			// Load source bioassay sets
			dc = TestUtil.getDbControl();
			Experiment experiment = PrepareTest.getExperiment(dc);
			String expression = Config.getFilterExpression();
			List<BioAssaySet> sources = new ArrayList<BioAssaySet>(
					Util.findAllBioAssaySets(dc, experiment, sourcePrefix + "%"));
			int numSources = sources.size();
			if (numSources == 0)
			{
				throw new ItemNotFoundException("BioAssaySet[name like " + sourcePrefix + "]");
			}
			dc.close();

			// Various export options
			// ch1 + ch2
			List<String> ch12 = Arrays.asList(new String[] {"intensity1", "intensity2"});
			// Reporter#externalId
			List<String> externalId = Arrays.asList(new String[] { "rep.externalId" });
			// All reporter properties
			List<String> allReporter = new ArrayList<String>();
			for (ExtendedProperty ep : ExtendedProperties.getProperties("ReporterData"))
			{
				allReporter.add("rep." + ep.getName());
			}
			// ch1 + ch2 + FGMean of both channels
			List<String> ch12FgMean = new ArrayList<String>();
			ch12FgMean.addAll(ch12);
			ch12FgMean.add("raw.ch1FgMean");
			ch12FgMean.add("raw.ch2FgMean");
			// ch1 + ch2 + all raw data properties
			List<String> ch12allRaw = new ArrayList<String>();
			ch12allRaw.addAll(ch12);
			for (RawDataProperty rdp : RawDataTypes.getRawDataType("genepix").getProperties())
			{
				ch12allRaw.add("raw." + rdp.getName());
			}

			for (BioAssaySet bas : sources)
			{
				exportBioAssaySet(bas, "S1", "basefile.serial", false, ch12, null, execute);
				exportBioAssaySet(bas, "M1", "basefile.matrix", false, ch12, null, execute);
				
				exportBioAssaySet(bas, "S2", "basefile.serial", false, ch12, externalId, execute);
				exportBioAssaySet(bas, "M2", "basefile.matrix", false, ch12, externalId, execute);
	
				exportBioAssaySet(bas, "S3", "basefile.serial", false, ch12, allReporter, execute);
				exportBioAssaySet(bas, "M3", "basefile.matrix", false, ch12, allReporter, execute);
				exportBioAssaySet(bas, "MeV", "mev.tdms", false, null, null, execute);
	
				exportBioAssaySet(bas, "S4", "basefile.serial", false, ch12FgMean, null, execute);
				exportBioAssaySet(bas, "M4", "basefile.matrix", false, ch12FgMean, null, execute);
	
				exportBioAssaySet(bas, "S5", "basefile.serial", false, ch12allRaw, null, execute);
				exportBioAssaySet(bas, "M5", "basefile.matrix", false, ch12allRaw, null, execute);
	
				exportBioAssaySet(bas, "S6", "basefile.serial", false, ch12allRaw, allReporter, execute);
				exportBioAssaySet(bas, "M6", "basefile.matrix", false, ch12allRaw, allReporter, execute);
			}
			TestUtil.write("---------------------------------------------\n");
		
			TestUtil.logout();
		}
		catch (Throwable t)
		{
			ok = false;
			t.printStackTrace();
		}
		finally
		{
			if (dc != null) dc.close();
		}
		TestUtil.write("++Leaving filter test (" + (ok ? "OK" : "FAILED") + ")\n");
		return ok;
	}

	/**
		Create a job for exporting the bioassay set.
	*/
	public static long exportBioAssaySet(BioAssaySet bioAssaySet, 
		String jobPrefix, String fileFormat, 
		boolean mergeOnReporter, List<String> fields, List<String> reporters,
		boolean execute)
		throws Exception
	{
		String name = bioAssaySet.getName();
		if (execute)
		{
			TestUtil.write("--Export " + jobPrefix + ": " + name + " to " + fileFormat + "\n");
		}
		DbControl dc = TestUtil.getDbControl();
		try
		{
			bioAssaySet = BioAssaySet.getById(dc, bioAssaySet.getId());
			PluginDefinition plugin = PluginDefinition.getByClassName(dc, "net.sf.basedb.plugins.BioAssaySetExporter");
			Job job = Job.getNew(dc, plugin, null, bioAssaySet.getExperiment());
			job.setName(jobPrefix + "." + name);
			job.setSendMessage(false);
			job.setRemoveJobWhenFinished(false);
			job.setParameterValue("bioAssaySet", 
					new ItemParameterType<BioAssaySet>(BioAssaySet.class, null), bioAssaySet);
			job.setParameterValue("fileformat", 
					new StringParameterType(255, null, true), fileFormat);
			job.setParameterValue("attachToBioAssaySet",	new BooleanParameterType(), false);
			job.setParameterValue("overwrite",	new BooleanParameterType(), true);
			job.setParameterValue("saveAs", new StringParameterType(255, null, true), "/" + jobPrefix + "." + fileFormat + ".txt");
			job.setParameterValue("geneAverages", new BooleanParameterType(), mergeOnReporter);
			job.setParameterValues("usedFields",  new StringParameterType(255, null, true, 0, 0, 0), fields);
			job.setParameterValues("usedColumns",  new StringParameterType(255, null, false, 0, 0, 0), reporters);
			dc.saveItem(job);
			dc.commit();
			long time = 0;
			if (execute)
			{
				time = System.currentTimeMillis();
				PluginUtil.executeJob(job);
				time = System.currentTimeMillis() - time;
				TestUtil.write(job.getName() + "\t" + Values.formatTime(time / 1000) + "\n");
			}
			else
			{
				TestUtil.write("Export job created: " + job.getName() + "\n");
			}
			return time;
			
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	
}
